#ifndef _COMMON_H
#define _COMMON_H

#include "tmwtypes.h"


/**** GENERAL PORPOUSE ******/
#define TRUE            1
#define FALSE           0
#define ERROR          -1
#define SUCCESS         0
#define CONNECT         1
#define HOLD            2
#define CANCEL          3
#define DISCONNECT      4
#define DA_RATE       100 /* Frequency of the Discarding Algorithm */

/****************************/


/********  macros for the fifos ********* */

#define RTF0         0   /* Data flowing from RTL to User Space */
#define RTF0_SIZE    (sizeof(int_T) * 1000)  
#define RTF1         1   /* Data flowing from User Space to RTL */
#define RTF1_SIZE    (sizeof(int_T) * 1000)  
 
/****************************************/


/***** commands sent through FIFO from krnl_main to us_main and vice versa *********/

#define SEND_DATA                 10
#define START_MODEL               11
#define STOP_MODEL                12
#define SET_PARAM                 13
#define PROCESS_CONNECT_MSG       14
#define PROCESS_CONNECT_MSG_RTL   15
#define UPLOAD_LOG_INFO_INIT      16
#define UPLOAD_INIT_TRIGGER       17
#define UPLOAD_ARM_TRIGGER        18
#define DO_DATA_LOG               19

/*******************************************************/


/************* Shared Memory ***********************/
/* SendData conatins the ring buffer where signal points are passed from the *
 * RT- proces to the network process.                                        
 * We don't use semaphores because even in multitasking (multithreaded) models,
 * this structure is only accessed by one thread (tid 0) in kernel space and the 
 * network process in user space.
 */

typedef struct SendDataStruc_tag{
	int           status;
	unsigned int  size ;
	unsigned int  hostNetRate;
	unsigned int  head;
	unsigned int  tail;
	char_T        *rt_buff;
	char_T        *net_buff;
} SendDataStruc;

typedef struct cmdHeader_tag{
	int type;
	int size;
} cmdHeader;


#define EXPAND_CONCAT(name1,name2) name1 ## name2
#define CONCAT(name1,name2) EXPAND_CONCAT(name1,name2)
#define RT_MODEL            CONCAT(MODEL,_rtModel)


/*****************
 * VERBOSE PRINT *
 *****************/

/* Messages. */
#ifdef VERBOSE
# define PRINT_M  printf       /* Print message. */
#else
# define PRINT_M  if(0) printf /* Do nothing. */
#endif
/* Warnings. */
#define VERBOSE_W
#ifdef VERBOSE_W
# define PRINT_W  printf       /* Print warnings. */
#else
# define PRINT_W  if(0) printf /* Do nothing. */
#endif
/* Errors. */
#define VERBOSE_E
#ifdef VERBOSE_E
# define PRINT_E fprintf       /* Print errors. */
#else
# define PRINT_E if(0) fprintf /* Do nothing. */
#endif


#endif _COMMON_H
