/* $Revision: 1.41 $
 * Copyright 1994-2002 The MathWorks, Inc.
 */
 
 #include <linux/malloc.h>

/* Shared memory header */
#include "mbuff.h"   

/* ST-RTL header */
#include "common.h"

/*Real Time Workshop headers*/
#include "tmwtypes.h"
#include "simstruc_types.h"

#include "ext_share.h"
#include "updown_util.h"
#include "dt_info.h"
#include "rtmodel.h"

/* rtl_printf() */
#include <rtl_printf.h>

#define printf rtl_printf


/* kmalloc() can allocate up to 128 kB of memory. This should be by far enough to
 * allocate all the structures. Only if thousands of signlas are selected, the required
 * amount of memory will be larger than 128kB. If this is the case, undefine the 
 * KMALLOC macro. 
 */ 
#define KMALLOC
#ifdef KMALLOC
# define mem_alloc(nBytes) kmalloc(nBytes, GFP_KERNEL)
# define mem_free(x)       if (x != NULL) kfree(x)
#else
# include <linux/vmalloc.h> 
# define mem_alloc(nBytes) vmalloc(nBytes)
# define mem_free(x)       if (x != NULL) vfree(x)
#endif 

int_T nBytesSec = 0;   /* Number of bytes per second to be sent to the simulink host. This       *
                        * value is used in ext_svr_rtl.c to work out the buffer size in SendData */
int_T nSPsec = 0;      /* Number of Simulink packets to be sent to the simulink host per second. *
                        * This value is used in ext_svr_rtl.c (DF_update()) with                 *
                        * SendData->hostNetRate to calculate the discarding Factor.              */


/**********************
 * External Variables *
 **********************/

/*NOTE: msg_shm, holds data in shared memory recieved from host and which must be pass down to the rl app.*/
static volatile char *msg_shm = NULL; 


/******************************************************************************
 * Parameter Download                                                         *
 ******************************************************************************/

#ifdef VERBOSE
/* Function: DType2Real_T ======================================================
 * Convert a built-in data type to a real_T value.  Return the real_T value
 * as well as a string (by reference) corresponding to the original data type.
 * If the data type is not recognized as a Simulink built-in data type, then
 * NULL is returned forthe dTypeName and 0 is returned for the dTypeValue.
 */
PRIVATE double DType2Double(
    const char              *vPtr,
    const int               dTypeIdx,
    const DataTypeTransInfo *dtInfo,
    const char              **dTypeName)
{
    real_T     outVal;
    char * const *dTypeNames = (char * const *) dtGetDataTypeNames(dtInfo);
    const char *thisName     = dTypeNames[dTypeIdx];

    *dTypeName = thisName;

    if (strcmp(thisName, "real_T") == 0) {
        outVal = (real_T) (((const real_T*)vPtr)[0]);
    } else if (strcmp(thisName, "real32_T") == 0) {
        outVal = (real_T) (((const real32_T*)vPtr)[0]);
    } else if (strcmp(thisName, "int8_T") == 0) {
        outVal = (real_T) (((const int8_T*)vPtr)[0]);
    } else if (strcmp(thisName, "uint8_T") == 0) {
        outVal = (real_T) (((const uint8_T*)vPtr)[0]);
    } else if (strcmp(thisName, "int16_T") == 0) {
        outVal = (real_T) (((const int16_T*)vPtr)[0]);
    } else if (strcmp(thisName, "uint16_T") == 0) {
        outVal = (real_T) (((const uint16_T*)vPtr)[0]);
    } else if (strcmp(thisName, "int32_T") == 0) {
        outVal = (real_T) (((const int32_T*)vPtr)[0]);
    } else if (strcmp(thisName, "uint32_T") == 0) {
        outVal = (real_T) (((const uint32_T*)vPtr)[0]);
    } else if (strcmp(thisName, "boolean_T") == 0) {
        outVal = (real_T) (((const boolean_T*)vPtr)[0]);
    } else {
        outVal    = 0;
        dTypeName = NULL;
    }
    return(outVal);
} /* end DType2Double */
#endif


/* Function: SetParam ==========================================================
 * Install new parameters.
 *
 * NOTE: pbuf looks like:
 *  [NPARAMS
 *   B S W DI DATA   % pVal 0
 *   B S W DI DATA   % pVal 1
 *   B S W DI DATA   % pVal N
 *  ]
 *
 *  No assumptions about the aligment of pbuf may be made.
 *
 *  where,
 *      B   : Data type transition index.  Note that on the target the data
 *            type transition index provides both the base address (B)
 *              of the transition as well as whether or not
 *              that section of the array contains complex elements.
 *      S   : starting offset of the data from the DT_TRANS_IDX
 *      W   : number of elements for this param
 *      DI  : index into rtw data type table (inUse idx)
 *      DATA: the param values (in target format).
 *
 *  and,
 *      All values, excluding DATA, are int32_T.
 */
PUBLIC void SetParam(RTWExtModeInfo  *ei, int_T size_shm)
{
    int        i;
    int32_T    nParams;
    const char *bufPtr    = NULL;  /* pbuf; */
    const int  B          = 0; /* index into dtype tran table (base address)  */
    const int  SI         = 1; /* starting index - wrt to base address        */
    const int  W          = 2; /* width of section (number of elements)       */
    const int  DI         = 3; /* index into data type tables                 */
    const int  tmpBufSize = sizeof(int32_T) * 4;
    int32_T    tmpBuf[4];

    const DataTypeTransInfo *dtInfo = rteiGetModelMappingInfo(ei);
    const DataTypeTransitionTable *dtTable = dtGetParamDataTypeTrans(dtInfo);
    const uint_T *dtSizes = dtGetDataTypeSizes(dtInfo);

     /* Use data allocated in shared memory */
     msg_shm = (char *) mbuff_alloc("msg_shm", size_shm);    
     bufPtr = (char *) msg_shm;

    /* unpack NPARAMS */
    (void)memcpy(&nParams, bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);
    
    PRINT_M("\nUpdating %d parameters....\n", nParams);

    /*
     * Unpack the data and install the new parameters.
     */
    for (i=0; i<nParams; i++) {
        int_T   elSize;
        int_T   nBytes;
        char_T  *start;
        char_T  *tranAddress;
        int_T   tranIsComplex;

        /* unpack B SI W DI */
        (void)memcpy(tmpBuf, bufPtr, tmpBufSize);
        bufPtr += tmpBufSize;

        /* 
         * Find starting address and size (nBytes) for this parameters 
         * section of memory.
         */
        tranAddress   = dtTransGetAddress(dtTable, tmpBuf[B]);
        tranIsComplex = dtTransGetComplexFlag(dtTable, tmpBuf[B]);

        elSize = dtSizes[tmpBuf[DI]] * (tranIsComplex ? 2 : 1);
        nBytes = tmpBuf[W] * elSize;
        start = tranAddress + (tmpBuf[SI] * elSize);

        /* Install the params. */
        (void)memcpy(start, bufPtr, nBytes);
        bufPtr += nBytes;

#ifdef VERBOSE
        /*
         * It is safe to assume that once the params are installed into
         * the param vector that they are properly aligned.  So we
         * do our verbosity print-out here.
         */
        {
            double     val;
            const char *dTypeName;
            const char *vPtr = (const char *)start;

            val = DType2Double(start, tmpBuf[DI], dtInfo, &dTypeName);
            PRINT_M("\n\tParam| "
                   "DT_Trans: %d, index: %d, nEls: %d, data type: [%s, %d]\n",
                   tmpBuf[B], tmpBuf[SI], tmpBuf[W],
                   (dTypeName != NULL) ? dTypeName : "",tmpBuf[DI]);

            if (!tranIsComplex) {
                int j;
                for (j=0; j<tmpBuf[W]; j++) {
                    val = DType2Double(vPtr, tmpBuf[DI], dtInfo, &dTypeName);
                    PRINT_M("\t\t%d/1000\n", (int) (val * 1000));
                    vPtr += dtSizes[tmpBuf[DI]];
                }
            } else {
                int j;
                for (j=0; j<tmpBuf[W]; j++) {
                    val = DType2Double(vPtr, tmpBuf[DI], dtInfo, &dTypeName);
                    PRINT_M("\t\t%d/1000 + ", (int) (val * 1000));
                    vPtr += dtSizes[tmpBuf[DI]];

                    val = DType2Double(vPtr, tmpBuf[DI], dtInfo, &dTypeName);
                    PRINT_M("\t\t%d/1000\n", (int) (val * 1000));
                    vPtr += dtSizes[tmpBuf[DI]];
                }
            }
        }
#endif
    }
/* free shared memory */
    mbuff_free("msg_shm", (void *) msg_shm);

} /* end SetParam */


/******************************************************************************
 * Parameter Upload                                                           *
 ******************************************************************************/
#include "upsup_public.h"
#define DUMP_MSG (0)


/*=============================================================================
 * Circular buffer stuff.
 *============================================================================*/
typedef struct BufMemList_tag {

#if ASSERTS_ON
    int_T  maxBufs; /* for debugging */
#endif

    int_T  nActiveBufs; /* num non-empty bufs                  */
    int_T  tids[NUMST]; /* tid associated with each active buf */

    BufMem *bufs; /* sections of each buffer to upload */
} BufMemList;


typedef struct CircularBuf_tag {
    int_T    empty;

    int_T    bufSize;
    char_T   *buf;
    
    char_T   *head;
    char_T   *tail;

    char_T   *newTail;

    struct {
        int_T count;
    } preTrig;
} CircularBuf;


/*==============================================================================
 * Trigger stuff.
 *============================================================================*/
typedef enum {
    TRIGGER_UNARMED,
    TRIGGER_HOLDING_OFF,
    TRIGGER_ARMED,
    TRIGGER_DELAYED,
    TRIGGER_FIRED,
    TRIGGER_TERMINATING,
    TRIGGER_SENDING_TERM_MSG
} TriggerState;


/*==============================================================================
 * General stuff.
 *============================================================================*/

/*
 * A BIOSection defines a contiguous section of the blockio structure.  Each
 * section consists of elements of the same data type and same complexity.
 * xxx 'start' should be a const pointer
 */
typedef struct BIOSection_tag {
    void   *start;
    int_T  nBytes;
} BIOSection;

/*
 * A BIOMap is an array of BIOSections.  Typically a map consists of all of the
 * sections of the blockio structure that are relevent to a given task.
 */
typedef struct BIOMap_tag {
    int32_T    nSections;
    BIOSection *sections;

    int_T nBytes;  /* total number of bytes in this map */
} BIOMap;


/*
 * Each system contains a table of BIOMap's (one for each tid).  If no data
 * uploading is being done for a given tid, the bioMap pointer is NULL.
 * The enableState field indicates whether the system in question is active.
 *
 * For a model with 5 total tids, the table looks like:
 *
 *               tid 0    tid 1    tid 2   tid 3   tid 4
 * bioMap      -------------------------------------------
 *             |  ptr  |  NULL  |  NULL  |  ptr  |  NULL |
 *             ----|-------------------------|------------
 *                 |                         |
 *                 v                         v
 *               BIOMap                    BIOMap
 *              for tid 0                 for tid 3
 */
typedef struct SysUploadTable_tag {
    int_T   *enableState;
    BIOMap  *bioMap[NUMST];
} SysUploadTable;


typedef struct TriggerInfo_tag {
    TriggerState            state;
    int_T                   tid;
    int32_T                 duration;
    int32_T                 holdOff;
    int32_T                 delay;
    int_T                   lookForRising;
    int_T                   lookForFalling;
    real_T                  level;
    int_T                   count;
    int_T                   overFlow;

    BIOMap                  trigSignals;
    real_T                  *oldTrigSigVals;
    int_T                   haveOldTrigSigVal;

    struct {
        int32_T    duration;
        int32_T    count;

        int_T      checkUnderFlow; /* ??? */
    } preTrig;
} TriggerInfo;


/*
 * The BdUploadInfo contains all information regarding data logging.
 */
struct BdUploadInfo_tag {
    int32_T        nSys;       /* # of sys's for which data logging is active */
    SysUploadTable *sysTables; /* array of SysUploadTables                    */

    CircularBuf  circBufs[NUMST];  /* circular buffers to store upload data */
    BufMemList   bufMemList;       /* A list of buffer memory that contains
                                    * data to upload */

    TriggerInfo  trigInfo;
};


/*==============================================================================
 * Global upload data.
 */

/*
 * Definitions that must match Simulink definitions.
 */

#define UPLOAD_RISING_TRIGGER                   ((int32_T)  0)
#define UPLOAD_FALLING_TRIGGER                  ((int32_T)  1)
#define UPLOAD_EITHER_RISING_OR_FALLING_TRIGGER ((int32_T)  2)

/*
 * Definitions.
 */
#define TRIGMODE_ONESHOT (-1)

static BdUploadInfo  uploadInfo;


/* Function ====================================================================
 * Initialize a BIOSection.
 */
PRIVATE void InitBioSection(
    RTWExtModeInfo  *ei,
    const int32_T   *buf,
    BIOSection      *section)   /* out */
{
    int_T                         elSize;
    int_T                         offset;
    int_T                         nBytes;
    char_T                        *tranAddress;
    int_T                         tranIsComplex;
    
    const DataTypeTransInfo *dtInfo = rteiGetModelMappingInfo(ei);
    const DataTypeTransitionTable *dtTable = dtGetBIODataTypeTrans(dtInfo);
    const uint_T *dtSizes = dtGetDataTypeSizes(dtInfo);

    const int BI = 0;  /* index into dtype tran table (base address)  */
    const int SI = 1;  /* starting index - wrt to base address        */
    const int W  = 2;  /* width of section (number of elements)       */
    const int DI = 3;  /* index into data type tables                 */

    tranAddress   = dtTransGetAddress(dtTable, buf[BI]);
    tranIsComplex = dtTransGetComplexFlag(dtTable, buf[BI]);

    elSize = dtSizes[buf[DI]] * (tranIsComplex ? 2 : 1);
    nBytes = buf[W] * elSize;
    offset = buf[SI] * elSize;

    section->start  = tranAddress + offset;
    section->nBytes = nBytes;
} /* end InitBioSection */


/* Function ====================================================================
 * Initialize a SysUploadTable.  The callerBufPtr points to the current place in
 * the EXT_SELECT_SIGNALS msg which should be the enableIdx field.  This
 * function moves the callerBufPtr to the next unread field of the message.
 */
PRIVATE boolean_T InitSysUploadTable(
    RTWExtModeInfo  *ei,
    SysUploadTable  *sysTable,
    const char      **callerBufPtr) /* in/out */
{
    int_T        i;
    int_T        nTids;
    const char_T *bufPtr = *callerBufPtr;
    boolean_T    error   = EXT_NO_ERROR;

    /*
     * Set pointer to enable mode.
     */
    {
        int32_T sysIdx;

        /* read sysIdx */
        (void)memcpy(&sysIdx, bufPtr, sizeof(int32_T));
        bufPtr += sizeof(int32_T);

        sysTable->enableState = rteiGetAddrOfSubSystemModeVector(ei,sysIdx);
    }

    /* In kernel space pointers not initialised are 0 (NULL). When freeing them,             *
     * 'uploadLogInfoTerm()' tries to access memory pointed by non-initialised pointers.     *
     * To prevent this from happening all pointers are initialised to NULL. If no address is *
     * given to them memory is not free                                                      */

     for (i=0; i<NUMST; i++)  {
         sysTable->bioMap[i] = NULL;
     }

    /*
     * Allocate/initialize each tid's bioMap.
     */

    /* ...read [nTids] */
    (void)memcpy(&nTids, bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);
    
    for (i=0; i<nTids; i++) {
        int32_T  tid;
        int32_T  section;
        BIOMap   *map;
        
        /* read tid */
        (void)memcpy(&tid, bufPtr, sizeof(int32_T));
        bufPtr += sizeof(int32_T);
        
        /* allocate BIOMap */
        assert(sysTable->bioMap[tid] == NULL);
        sysTable->bioMap[tid] = (BIOMap *)mem_alloc(sizeof(BIOMap));
        if (sysTable->bioMap[tid] == NULL) {
             PRINT_W("\n!!!! Memory could not be allocated in kernel space. !!!\n");
             error = EXT_ERROR; goto EXIT_POINT;       
        }
        memset(sysTable->bioMap[tid], 0, sizeof(BIOMap));
        map = sysTable->bioMap[tid];
        
        /* read nSections */
        (void)memcpy(&map->nSections, bufPtr, sizeof(int32_T));
        bufPtr += sizeof(int32_T);

        /* Allocate the blockio sections. */
        assert(map->sections == NULL);
        map->sections = (BIOSection *)mem_alloc(map->nSections * sizeof(BIOSection));
        if (map->sections == NULL) {
         PRINT_W("\n!!!! Memory could not be allocated in kernel space. \
                   \n\tThis problem could be fixed by doing #undef KMALLOC \
                       in updown.c (line 32).  !!!\n");
            error = EXT_ERROR; goto EXIT_POINT;
        }
        memset(map->sections, 0, map->nSections * sizeof(BIOSection));

        /*
         * Init the BIOSections.
         */
        for (section=0; section<map->nSections; section++) {
            int32_T    tmpBuf[4];
            BIOSection *bioSection = &map->sections[section];

            /* read [B S W DI] */
            (void)memcpy(&tmpBuf, bufPtr, sizeof(int32_T)*4);
            bufPtr += (sizeof(int32_T) * 4);

            InitBioSection(ei, tmpBuf, bioSection);

            /* keep track of total number of bytes in this map */
            map->nBytes += bioSection->nBytes;
        }
    }
    
EXIT_POINT:
    *callerBufPtr = bufPtr;
    return(error);
} /* end InitSysUploadTable */


/* Function ====================================================================
 * Initialize circular buffer fields and allocate required memory.
 */
PRIVATE boolean_T UploadBufInit(CircularBuf *circBuf, int_T size)
{
    boolean_T error = NO_ERR;

    assert(circBuf->buf == NULL);

    circBuf->empty = TRUE;
    if (size > 0) {
        assert(circBuf->buf == NULL);
        circBuf->buf = (char_T *)mem_alloc(size);
        if (circBuf->buf == NULL) {
         PRINT_W("\n!!!! Memory could not be allocated in kernel space. \
                   \n\tThis problem could be fixed by doing #undef KMALLOC \
                       in updown.c (line 32).  !!!\n");
            error = EXT_ERROR; goto EXIT_POINT;
        }
    } else {
        circBuf->buf = NULL;
    }
    circBuf->bufSize = size;
    
    circBuf->head = circBuf->buf;
    circBuf->tail = circBuf->buf;

    circBuf->newTail = NULL;

EXIT_POINT:
    return(error);
} /* end UploadBufInit */


/* Function ====================================================================
 * Reset all data associated with a circular buffers.
 */
PRIVATE void UploadBufReset(CircularBuf *circBuf)
{
    circBuf->empty   = TRUE;
    circBuf->bufSize = 0;
    circBuf->buf     = NULL;

    circBuf->head    = NULL;
    circBuf->tail    = NULL;

    circBuf->newTail = NULL;
} /* end UploadBufReset */


/* Function ====================================================================
 * Reset fields of the uploadinfo struct.
 */
PUBLIC void UploadLogInfoReset(void)
{
    int i;

    /* sysUploadTable */
    uploadInfo.nSys      = 0;
    uploadInfo.sysTables = NULL;

    /* circular bufs and bufMemLists */
    for (i=0; i<NUMST; i++) {
        UploadBufReset(&uploadInfo.circBufs[i]);
    }

    uploadInfo.bufMemList.bufs = NULL;

    /*
     * Reset trigger info.
     */
    uploadInfo.trigInfo.state          = TRIGGER_UNARMED;
    uploadInfo.trigInfo.duration       = 0;
    uploadInfo.trigInfo.holdOff        = 0;
    uploadInfo.trigInfo.delay          = 0;
    uploadInfo.trigInfo.lookForRising  = TRUE;
    uploadInfo.trigInfo.lookForFalling = FALSE;
    uploadInfo.trigInfo.level          = 0.0;
    uploadInfo.trigInfo.count          = 0;
    uploadInfo.trigInfo.overFlow       = FALSE;

    uploadInfo.trigInfo.trigSignals.nSections = 0;
    uploadInfo.trigInfo.trigSignals.sections  = NULL;
    uploadInfo.trigInfo.trigSignals.nBytes    = 0;
    
    uploadInfo.trigInfo.oldTrigSigVals    = NULL;
    uploadInfo.trigInfo.haveOldTrigSigVal = FALSE;

    uploadInfo.trigInfo.preTrig.duration       = 0;
    uploadInfo.trigInfo.preTrig.count          = 0;
    uploadInfo.trigInfo.preTrig.checkUnderFlow = FALSE;
} /* end UploadLogInfoReset */


/* Function ====================================================================
 * Free all dynamically allocated fields of the trigInfo structure.
 */
PRIVATE void UploadDestroyTrigger(void)
{
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;
    BIOMap      *map      = &trigInfo->trigSignals;

    mem_free(map->sections);
    map->sections = NULL;
    
    mem_free(trigInfo->oldTrigSigVals);
    trigInfo->oldTrigSigVals = NULL;

} /* end UploadDestroyTrigger */


/* Function ====================================================================
 * Destroy all data associated with data logging.  Fields are re-initialized
 * and pointers NULL'ed out by UploadBufReset() and UploadLogInfoReset().
 */
PUBLIC void UploadLogInfoTerm(void)
{
    int_T i;

    /*
     * Free fields of the sysUpload tables and then the table itself.
     */
    for (i=0; i<uploadInfo.nSys; i++) {
        int_T   tid;
        BIOMap  **bioMap = uploadInfo.sysTables[i].bioMap;
        
        for (tid=0; tid<NUMST; tid++) {
            if (bioMap[tid] != NULL) {
                /* Free fields of bioMap. */
                mem_free(bioMap[tid]->sections);
                /* Free the bioMap. */
                mem_free(bioMap[tid]);
            }
        }
    }
    mem_free(uploadInfo.sysTables);

    /* Free ciruclar buf fields and bufMemLists. */
    for (i=0; i<NUMST; i++) {
        mem_free(uploadInfo.circBufs[i].buf);
    }

    mem_free(uploadInfo.bufMemList.bufs);
    
    /*
     * Free trigger info.
     */
    UploadDestroyTrigger();

    /*
     * Reset all vals to initial value and NULL out pointers.
     */
    UploadLogInfoReset();

} /* end UploadLogInfoTerm */

/* Function ====================================================================
 * Prepare for final flush of buffers.  This involves setting the trigger
 * state to appropriate values.
 */
PUBLIC void UploadPrepareForFinalFlush(void)
{
    switch(uploadInfo.trigInfo.state) {
    case TRIGGER_FIRED:
    case TRIGGER_TERMINATING:
        /*
         * 1) set trig state to "terminating" so that the eventual call to
         *    UploadBufGetData knows to add the terminator flag to
         *    the data stream.
         * 2) set trig state to "oneshot" to prevent re-arming
         */
        uploadInfo.trigInfo.state   = TRIGGER_TERMINATING; /* 1 */
        uploadInfo.trigInfo.holdOff = TRIGMODE_ONESHOT;    /* 2 */
        break;

    case TRIGGER_UNARMED:
    case TRIGGER_HOLDING_OFF:
    case TRIGGER_ARMED:
    case TRIGGER_DELAYED:
    case TRIGGER_SENDING_TERM_MSG:
        /* do nothing */
        break;
    }

#ifdef VXWORKS
    /* Let upload server run to ensure that term msg is sent to host. One
       semGive() is for the background task and the other is for the explicit
       call to rt_UploadServerWork() in DisconnectFromHost(). */
    semGive(uploadSem);
    semGive(uploadSem);
#endif
	
} /* end UploadPrepareForFinalFlush */


/* Function ====================================================================
 * Shut down data logging and do cleanup.
 */
PUBLIC void UploadShutDown(void)
{
    UploadLogInfoTerm();

} /* end UploadShutDown */


/* Function ====================================================================
 * Initialize data uploading by processing the EXT_SELECT_SIGNALS message
 * (which is passed in).  Return the error status.  See DumpSelectSignalMsg()
 * for a detailed description of the message.
 */
PUBLIC boolean_T UploadLogInfoInit(RT_MODEL *S,
                                   int_T size_shm)
{
    int         nActiveTids;
    int_T       i;
    boolean_T   error   = EXT_NO_ERROR;
    const char  *bufPtr = NULL;
    RTWExtModeInfo  *ei = rtmGetRTWExtModeInfo(S);

     /* Use data allocated in shared memory */
     msg_shm = (char *) mbuff_alloc("msg_shm", size_shm);    
     if (msg_shm == NULL) {
        PRINT_W("\n!!! 'msg_shm' could not be allocated in shared memory!!!\n");
        error = EXT_ERROR;  goto EXIT_POINT;
     }

     /* because the address where bufPtr changes we cannot free memory using bufPtr. 
      * Therefore we need msg_shm,which constantly points to the base address of the shared memory */
     bufPtr = (const char *) msg_shm; 

    /*
     * Don't init data logging while it is already initialized.  In most cases,
     * this is prevented at a higer level (GUI disabled, etc).
     */
    /* if (uploadInfo.sysTables != NULL) return(EXT_ERROR); */

    /* nSys */
    (void)memcpy(&uploadInfo.nSys, bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);
    
    assert(uploadInfo.sysTables == NULL);
    uploadInfo.sysTables = (SysUploadTable *)
                           mem_alloc(uploadInfo.nSys * sizeof(SysUploadTable));
    if (uploadInfo.sysTables == NULL) {
         PRINT_W("\n!!!! Memory could not be allocated in kernel space. \
                   \n\tThis problem could be fixed by doing #undef KMALLOC \
                       in updown.c (line 32).  !!!\n");
        error = EXT_ERROR; goto EXIT_POINT;
    }
    memset(uploadInfo.sysTables, 0, uploadInfo.nSys * sizeof(SysUploadTable));

    /*
     * Init each system table.
     */
    for (i=0; i<uploadInfo.nSys; i++) {
        error = InitSysUploadTable(ei, &uploadInfo.sysTables[i], &bufPtr);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
    }

    /*
     * Allocate the circular buffers.
     * RTW allocates NUMSTs ring buffers. As we use shared memory, data is no longer permantely    *
     * sotred in those buffers. However they are still needed, since the output of every cycle is  *
     * copied in the bufferes. Afterwards that data is transferred to the shared memory. The size  *
     * of the ring buffers must be enough to store data for at least one sampling time. However,   *
     * memory is allocated for two sampling time, leaving one as a margin                          * 

     * The size is given by the following expression:                                              *
     * size = [((# samples * GSampleTime / BSampleTime) + 1) * (72 + (nSections - 1) * 16)]        *
     * where GSampleTime is the 'fixed step size' and BSampleTime is the  tid's sample time        *
     * In our case, samples is 2                                                                   */

    /* Reset the number of bytes to be sent per second  */
    nBytesSec = 0;
    nActiveTids = 0;
    for (i=0; i<NUMST; i++) {
        int32_T size;

        /* Original size. */
        (void)memcpy(&size, bufPtr, sizeof(int32_T));
        bufPtr += sizeof(int32_T);
        
        if (size > 0) {
           if (uploadInfo.sysTables->bioMap[i]) {
                real_T BSampleTime = rtmGetSampleTime(S, i);      /* block's sample time.  */
                /*
                 * RTW allocates NUMSTs upload buffers. Simulink packets are temporarily stored before copied  *
                 * onto shared memory, which is done within the same cycle. Therefore the RTW upload buffers   *
                 * only store one packet at a time. Its size can therefore be reduced to optimize memory usage.*
                 * The consequent buffer sizes equals to the Simulink packets size for any given tid           *
                 */
	          size = (int32_T) uploadInfo.sysTables->bioMap[i]->nBytes + 28; /* 28 byte header */

                /* nBytesSec: is the total number of bytes per second to be sent to the Simulink host */
                nBytesSec += (int_T) (1/BSampleTime + 1) * size;
                nSPsec    += 1/BSampleTime + 1;
            } else {
		    /* RMG: If the parameter duration in "Exteranl Mode control panel - > signal & triggering 
                 * is maximum (2147483647) and a continuos state has to be uploaded, the variable
		     * size equals to this value for the continuous (sample rate 0) and the base sample rate.
		     * Yet, uploadInfo.sysTables->bioMap[i] is non-NULL only for the latter. When recalculating
                 * the size (above) this would yield a segmentation error. Hence the condition.MathWorks bug again!!!.	
		     */
		    size = 0;
		}
	  }
        /* Allocate the circular buffers. */
        if (UploadBufInit(&uploadInfo.circBufs[i], size) == EXT_ERROR) {
           error = ERROR;  goto EXIT_POINT;
        }
        nActiveTids += (size != 0);
    }

    /*
     * Initialize/Allocate the bufMemLists - these are used by
     * ext_svr to pull the appropriate data out of the buffers and send it
     * to the host.
     */
#if ASSERTS_ON
    uploadInfo.bufMemList.maxBufs = nActiveTids;
#endif
    uploadInfo.bufMemList.nActiveBufs = 0;

    assert(uploadInfo.bufMemList.bufs == NULL);
    uploadInfo.bufMemList.bufs = (BufMem *)mem_alloc(nActiveTids*sizeof(BufMem));
    if (uploadInfo.bufMemList.bufs == NULL) {
         PRINT_W("\n!!!! Memory could not be allocated in kernel space. \
                   \n\tThis problem could be fixed by doing #undef KMALLOC \
                       in updown.c (line 32).  !!!\n");
        error = EXT_ERROR; goto EXIT_POINT;
    }
    memset(uploadInfo.bufMemList.bufs, 0, nActiveTids * sizeof(BufMem));

    /* free shared memory */
    mbuff_free("msg_shm", (void *) msg_shm);

EXIT_POINT:
    /* free shared memory */
    mbuff_free("msg_shm", (void *) msg_shm);

    if (error != EXT_NO_ERROR) {
        UploadLogInfoTerm();
    }
    return(error);
} /* end UploadLogInfoInit */


/* Function ====================================================================
 * Initialize and configure the trigger attributes.  See DumpSelectTriggerMsg()
 * for a detailed description of the message.
 */
PUBLIC boolean_T UploadInitTrigger(RTWExtModeInfo *ei,
                                   int_T size_shm)
{
    int_T       nSections;
    int32_T     i32_tid;
    int32_T     direction;
    boolean_T   error     = EXT_NO_ERROR;
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;
    volatile char  *bufPtr   = NULL; /* msg; */

    msg_shm = (char *) mbuff_alloc("msg_shm", size_shm);    
    if (msg_shm == NULL) {
      PRINT_W("\n!!! 'msg_shm' could not be allocated in shared memory!!!\n");
      error = ERROR; 
      goto EXIT_POINT;
    }
     /* because the address where bufPtr changes we cannot free memory using bufPtr. 
      * Therefore we need msg_shm,which constantly points to the base address of the shared memory */
     bufPtr = msg_shm; 

    /* tid, duration, holdOff, delay and nsections */
    (void)memcpy(&i32_tid, (char *) bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);
    trigInfo->tid = (int_T)i32_tid;

    (void)memcpy(&trigInfo->duration, (char *) bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);

    (void)memcpy(&trigInfo->holdOff, (char *)bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);

    (void)memcpy(&trigInfo->delay, (char *) bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);

    (void)memcpy(&trigInfo->trigSignals.nSections, (char *) bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);
    
    nSections = trigInfo->trigSignals.nSections;

    /*
     * Init the BIOSections - if the trigger is signal based.
     */
    if (nSections > 0) {
        /* trigger is signal based */
        int    section;
        BIOMap *map = &trigInfo->trigSignals;

        assert(map->nBytes == 0);
        assert(map->sections == NULL);
        map->sections = (BIOSection *)mem_alloc(nSections * sizeof(BIOSection));
        if (map->sections == NULL) {
         PRINT_W("\n!!!! Memory could not be allocated in kernel space. \
                   \n\tThis problem could be fixed by doing #undef KMALLOC \
                       in updown.c (line 32).  !!!\n");
            error = EXT_ERROR; goto EXIT_POINT;
        }
        memset(map->sections, 0, nSections * sizeof(BIOSection));
            
        for (section=0; section<map->nSections; section++) {
            int32_T    tmpBuf[4];
            BIOSection *bioSection = &map->sections[section];

            /* read [B S W DI] */
            (void)memcpy(&tmpBuf, (char *) bufPtr, sizeof(int32_T)*4);
            bufPtr += (sizeof(int32_T) * 4);
            
            InitBioSection(ei, tmpBuf, bioSection);

            /* keep track of total number of bytes in this map */
            map->nBytes += bioSection->nBytes;
        }

        /*
         * Allocate space to hold the old values of the trigger signals.  Note
         * that trigger signals are gauranteed to be of type SL_DOUBLE (real_T)
         * and noncomplex.
         */
        assert(trigInfo->oldTrigSigVals == NULL);
        assert(trigInfo->oldTrigSigVals == NULL);
        trigInfo->oldTrigSigVals = (real_T *)mem_alloc(map->nBytes);
        if (trigInfo->oldTrigSigVals == NULL) {
            PRINT_W("\n!!!! Memory could not be allocated in kernel space. !!!\n");
            error = EXT_ERROR; goto EXIT_POINT;
        }
        memset(trigInfo->oldTrigSigVals, 0, map->nBytes);
    }
            
    /* Direction. */
    (void)memcpy(&direction, (char *) bufPtr, sizeof(int32_T));
    bufPtr += sizeof(int32_T);

    trigInfo->lookForRising  = 
        ((direction == UPLOAD_RISING_TRIGGER)  || 
         (direction == UPLOAD_EITHER_RISING_OR_FALLING_TRIGGER));

    trigInfo->lookForFalling = 
        ((direction == UPLOAD_FALLING_TRIGGER) || 
         (direction == UPLOAD_EITHER_RISING_OR_FALLING_TRIGGER));
    
    /* level */
    (void)memcpy(&trigInfo->level, (char *) bufPtr, sizeof(real_T));
    bufPtr += sizeof(real_T);

    /*
     * Convert delay to pre-trigger duration.
     */
    if (trigInfo->delay < 0) {
        trigInfo->preTrig.duration = -trigInfo->delay;
        trigInfo->delay            = 0;
    } else {
        trigInfo->preTrig.duration = 0;
    }

EXIT_POINT:
    if (error != EXT_NO_ERROR) {
        UploadDestroyTrigger();
    }
    return(error);
} /* end UploadInitTrigger */


/* Function ====================================================================
 * Arm the trigger.
 */
PUBLIC void UploadArmTrigger(void)
{
    int_T tid;

    assert((uploadInfo.trigInfo.state == TRIGGER_UNARMED) ||
           (uploadInfo.trigInfo.state == TRIGGER_HOLDING_OFF));

    /*
     * Re-initialize.
     */
    uploadInfo.trigInfo.overFlow = FALSE;
    for (tid=0; tid<NUMST; tid++) {
        CircularBuf *circBuf = &uploadInfo.circBufs[tid];
        if (circBuf->bufSize > 0) {
            circBuf->head = circBuf->buf;
            circBuf->tail = circBuf->buf;

            circBuf->newTail = NULL;
            circBuf->empty   = TRUE;
        }
    }

    /* 
     * Re-initialize trigger fields.
     */
    uploadInfo.trigInfo.count             = 0;
    uploadInfo.trigInfo.haveOldTrigSigVal = FALSE;

    /* 
     * Reset pre-trig counts for normal mode.
     */
    uploadInfo.trigInfo.preTrig.count = 0;

    /* 
     * Re-arm after all initialization.  Make sure that trigInfo.state is
     * set last since this routine may be interupted.
     */
    uploadInfo.trigInfo.state = TRIGGER_ARMED;

} /* end UploadArmTrigger */


/* Function ====================================================================
 * Terminate this data logging session by destroying the uploadInfo and
 * setting the trigger backed to the unarmed state.
 */
PUBLIC void EndLoggingSession(void)
{
    uploadInfo.trigInfo.state = TRIGGER_UNARMED;
    UploadLogInfoTerm();
} /* end UploadEndLoggingSession */


/* Function ====================================================================
 * Cancel this data logging session.
 */
PUBLIC void UploadCancelLogging(void)
{
    switch(uploadInfo.trigInfo.state) {

    case TRIGGER_UNARMED:
        break;

    case TRIGGER_HOLDING_OFF:
    case TRIGGER_ARMED:
    case TRIGGER_DELAYED:
    case TRIGGER_FIRED:
        /*
         * Move to TRIGGER_TERMINATING and ensure that we are no longer in
         * "normal" mode (TRIGMODE_NORMAL) so that the trigger does not get
         * re-armed.
         */
        uploadInfo.trigInfo.holdOff = TRIGMODE_ONESHOT;
        uploadInfo.trigInfo.state   = TRIGGER_TERMINATING;
        break;
    
    case TRIGGER_TERMINATING:
    case TRIGGER_SENDING_TERM_MSG:
        /*
         * Ensure that we are no longer in "normal" mode (TRIGMODE_NORMAL) so
         * that the trigger does not get re-armed.
         */
        uploadInfo.trigInfo.holdOff = TRIGMODE_ONESHOT;
        break;
    }
} /* end UploadCancelLogEvent */


/* Function ====================================================================
 * Called by ext_svr (background task), in order to perform tasks that need
 * to be done after each time that data has been sent to the host.  This
 * includes:
 *
 * o move the tail for the specified buffer forward
 * o detect the end of a data logging event so that the trigger state can
 *   be either set to unarmed (for one shot) or backed to armed (for normal
 *   mode).
 * 
 * NOTE:  UploadBufGetData and UploadBufMoveTail must be called in pairs where the
 *        UploadBufGetData call precedes the UploadBufMoveTail call.
 */
PUBLIC void UploadBufDataSent(const int_T tid)
{
    CircularBuf *circBuf = &uploadInfo.circBufs[tid];
    
    /* 
     * Move the tail forward.  Since we are moving the tail forward, we know that
     * head == tail represents an empty buffer and not a full buffer.
     */
    circBuf->tail = circBuf->newTail;
    circBuf->empty = (circBuf->tail == circBuf->head);

    /*
     * See if this is the end of the data logging event.
     */
    if (uploadInfo.trigInfo.state == TRIGGER_SENDING_TERM_MSG) {
        if (uploadInfo.trigInfo.holdOff == TRIGMODE_ONESHOT) {
            /* The memory cannot be freed from here because this is run within the RT-thread *
             * This is done  in ext_svr_us/rt_UploadServerWork                               */
            /* EndLoggingSession(); */
        } else {
            uploadInfo.trigInfo.count = 0;
            uploadInfo.trigInfo.state = TRIGGER_HOLDING_OFF;
        }
    }
} /* end UploadBufDataSent */


/*
 * Macro =======================================================================
 * Move the tail of a circular buffer forward by one time step - accounting for
 * wrapping.
 */
#define MOVE_TAIL_ONESTEP(circBuf, end)                            \
{                                                                  \
    int  nBytesPassedEnd;                                          \
    int  nBytesInStep;                                             \
                                                                   \
    (void)memcpy(&nBytesInStep, (circBuf)->tail, sizeof(int32_T)); \
    assert(nBytesInStep > 0);                                      \
    (circBuf)->tail += (nBytesInStep);                             \
    nBytesPassedEnd = (circBuf)->tail - (end);                     \
    if (nBytesPassedEnd >= 0) {                                    \
        (circBuf)->tail = (circBuf)->buf + nBytesPassedEnd;        \
    }                                                              \
} /* end MOVE_TAIL_ONESTEP */


/*
 * Macro =======================================================================
 * Copy data into the circular buffer.
 */
#define CIRCBUF_COPY_DATA(bufMem, data)                         \
{                                                               \
    (void)memcpy((bufMem).section1, (data), (bufMem).nBytes1);  \
    if ((bufMem).section2 != NULL) {                            \
        char *tmp = ((char *)(data)) + (bufMem).nBytes1;        \
        (void)memcpy((bufMem).section2, tmp, (bufMem).nBytes2); \
    }                                                           \
} /* end CIRCBUF_COPY_DATA */


/* Function ====================================================================
 * Assign sections in the circular buffer for the requested number of bytes
 * (i.e., fill in the bufMem struct).  If there is no room in the circular
 * buffer return an overflow error.
 *
 * NOTE: Do not move the CircularBuffers head forward in this function!  
 *       Only move the tmpHead forward.  The actual head is not advanced
 *       until the entire time point is successfully copied into the buffer.
 *
 *       This function modifies tmpHead to point at the next available 
 *       location.
 *
 *       It is possible for tmpHead to equal the tail upon entry to this
 *       function.  This does not necessarily mean that the buffer is
 *       empty (unwrapped).  It could also mean that the buffer is exactly
 *       full (this is considered as wrapped).
 */
PRIVATE boolean_T UploadBufAssignMem(
    CircularBuf  *circBuf,
    int_T        nBytesToAdd,
    char         **tmpHead,   /* in-out */
    BufMem       *bufMem)     /* out */
{
    int_T       nBytesLeft;
    boolean_T   overFlow  = FALSE;
    char        *end      = circBuf->buf + circBuf->bufSize; /* 1 passed end */

    if ((*tmpHead > circBuf->tail) || circBuf->empty) {
        /* buffer not wrapped */
        nBytesLeft = (end - *tmpHead) + (circBuf->tail - circBuf->buf);

        if (nBytesLeft < nBytesToAdd) {
            overFlow = TRUE;
            goto EXIT_POINT;
        }

        if ((*tmpHead + nBytesToAdd) < end) {
            /* still not wrapped */
            bufMem->nBytes1  = nBytesToAdd;
            bufMem->section1 = *tmpHead;

            bufMem->nBytes2  = 0;
            bufMem->section2 = NULL;

            *tmpHead += nBytesToAdd;
        } else {
            /* now we're wrapped */
            bufMem->nBytes1  = end - *tmpHead;
            bufMem->section1 = *tmpHead;

            bufMem->nBytes2  = nBytesToAdd - bufMem->nBytes1;
            bufMem->section2 = (bufMem->nBytes2 > 0) ? circBuf->buf : NULL;

            *tmpHead = circBuf->buf + bufMem->nBytes2;
        }  
    } else {
        /* wrapped */
        nBytesLeft = circBuf->tail - *tmpHead;
        if (nBytesLeft < nBytesToAdd) {
            overFlow = TRUE;
            goto EXIT_POINT;
        }

        bufMem->nBytes1  = nBytesToAdd;
        bufMem->section1 = *tmpHead;

        bufMem->nBytes2  = 0;
        bufMem->section2 = NULL;

        *tmpHead += nBytesToAdd;
    }
    
EXIT_POINT:
    return(overFlow);
} /* end UploadBufAssignMem */


/* Function ====================================================================
 * Check the trigger signals for crossings.  Return true if a trigger event is
 * encountered.  It is assumed that the trigger signals are real_T.
 */
PRIVATE boolean_T UploadCheckTriggerSignals(void)
{
    int         i;
    TriggerInfo *trigInfo       = &uploadInfo.trigInfo;
    real_T      *oldTrigSigVals = trigInfo->oldTrigSigVals;
    real_T      *oldSigPtr      = oldTrigSigVals;
       
    for (i=0; i<trigInfo->trigSignals.nSections; i++) {
        BIOSection *section = &trigInfo->trigSignals.sections[i];
        int_T      nEls     = section->nBytes / sizeof(real_T); /* xxx cache? */

        /*
         * If we have a previous signal value to check, then see if we had
         * a crossing.
         */
        if (trigInfo->haveOldTrigSigVal) {
            int_T   j;
            real_T  level   = trigInfo->level;
            real_T  *rStart = (real_T *)section->start; /* gauranteed by host */
            
            for (j=0; j<nEls; j++) {
                if (trigInfo->lookForRising && 
                    (((rStart[j] >= level) && (oldSigPtr[j] <  level)) ||
                     ((rStart[j] >  level) && (oldSigPtr[j] == level)))) {
                    return(TRUE);
                }
                if (trigInfo->lookForFalling &&
                    (((rStart[j] < level)  && (oldSigPtr[j] >= level)) ||
                     ((rStart[j] == level) && (oldSigPtr[j] >  level)))) {
                    return(TRUE);
                }
            }
        }

        /*
         * Update old signal values.
         */
        (void)memcpy(oldSigPtr, section->start, section->nBytes);
        oldSigPtr += section->nBytes;
    }
    assert(oldTrigSigVals + trigInfo->trigSignals.nBytes == oldSigPtr);
    trigInfo->haveOldTrigSigVal = TRUE;
    return(FALSE);
} /* end UploadCheckTriggerSignals */


/* Function ====================================================================
 * If the trigger is in the TRIGGER_FIRED state or we are collecting data for
 * pretriggering, add data, for each tid with a hit, to the upload buffers.  
 * This function is called from within the appropriate task, once per sample
 * hit.
 *
 * The format of the message that is sent to the host is as follows:
 *
 * definitions:
 *      nBytes - total number of target bytes for this message (including the
 *               nBytes field).  nBytes worth of data represents 1 full time
 *               step of the target simulation.
 *
 *      msgType - A qualifier indicating any special action that needs to be
 *                taken (e.g., a termination flag following the last data point,
 *                or a flag indicating that it is the first data point after
 *                a trigger event).
 *
 *      nSys - The number of systems for which this message contains data.
 *
 *      tid - The tid with which this data is associated.
 *
 *      t - simulation time
 *
 *      sysId - The index into the BdUploadInfo.sysTables array so that we can
 *              map the target data back to the appropriate sytstem.  This is
 *              NOT the descendent system index!
 *
 *      data - the target simulation data (in target format)
 *
 * The message looks like:
 * [nBytes msgType nSys tid t sysId [data] sysId [data]...]
 *     |                     | |         | |          |
 *     ----------------------- ----------- ------------
 *          msg header          sys data     sys data
 *
 * Ints are int32_T.
 */
 
PUBLIC void UploadBufAddTimePoint(const int_T tid, real_T taskTime)
{
    int_T       preTrig;
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;
    CircularBuf *circBuf  = &uploadInfo.circBufs[tid];
    int_T       overFlow  = FALSE;

     /* If MULTITASKING is defined, there is a continuouns block and the 
      * trigger is fired automatically (manual source)then 
      * trigInfo->state = TRIGGER_FIRED will never be true because 
      * tid == trigInfo->tid is always FASLSE, since tid = 1 and trigInfo->tid = 0;
      * THIS IS A MISTAKE FROM MATHWORKS. To solve this problem, if tid = 1 we make it be 0;
      */
#ifdef MULTITASKING
    if (trigInfo->tid == 0)
        trigInfo->tid ++;
#endif

    /*
     * Check for transitions from the TRIGGER_ARMED state to either the
     * TRIGGER_FIRED_STATE or the TRIGGER_DELAYED state.  We only do this
     * if it is a sample hit for the trigger signal.  Note that this
     * is the only place in the whole world that the trigger state can
     * move from TRIGGER_ARMED_STATE to TRIGGER_DELAYED or TRIGGER_FIRED.
     */
    if ((tid == trigInfo->tid) && (trigInfo->state == TRIGGER_ARMED)) {
        if ((trigInfo->trigSignals.nSections == 0) || /* manual trigger       */
            (UploadCheckTriggerSignals())) {          /* trig signal crossing */
    
            if (trigInfo->delay == 0) {
                trigInfo->state = TRIGGER_FIRED;
                trigInfo->count = trigInfo->preTrig.count; /* 0 unless pretrig */
            } else {
                trigInfo->state = TRIGGER_DELAYED;
                assert(trigInfo->count == 0);

                /* We will be skipping this step, so the delay count is 1. */
                trigInfo->count = 1;
            }
        }
    }

    preTrig = (trigInfo->state == TRIGGER_ARMED) &&
              (trigInfo->preTrig.duration > 0);
    
    /*
     * Handle adding data to the collection buffers - if needed.
     */
    if (((trigInfo->state == TRIGGER_FIRED) || preTrig) &&
        circBuf->bufSize != 0) { /* bufSize == 0 means no signals in this tid */

        int_T       i;
        BufMem      bufMem;
        BufMem      msgStart;
        int_T       size;
        char_T      *tmpHead    = circBuf->head;

        const int_T NBYTES_IDX   = 0;
        const int_T MSG_TYPE_IDX = 1;
        const int_T NSYS_IDX     = 2;
        const int_T TID_IDX      = 3;
        int32_T     intHdr[4]    = {0,0,0,0};

        if (preTrig && (trigInfo->preTrig.count==trigInfo->preTrig.duration)) {
            /* Advance the tail (we don't need the oldest point anymore). */
            char *end = circBuf->buf + circBuf->bufSize;
            MOVE_TAIL_ONESTEP(circBuf, end);
            trigInfo->preTrig.count--;
        }

        /*
         * Save some space for the 4 integer values that make up the message
         * header: [nBytes msgType nSys tid].  The values are filled in later.
         */
        size = 4*sizeof(int32_T);
        overFlow = UploadBufAssignMem(circBuf, size, &tmpHead, &msgStart);
        if (overFlow) goto EXIT_POINT;
        intHdr[NBYTES_IDX] += size;

        /* time */
        overFlow = UploadBufAssignMem(
            circBuf, sizeof(real_T), &tmpHead, &bufMem);
        if (overFlow) goto EXIT_POINT;
        intHdr[NBYTES_IDX] += sizeof(real_T);
        
        CIRCBUF_COPY_DATA(bufMem, &taskTime);

        /*
         * Check each system for a BIOMap. 
         */
        for (i=0; i<uploadInfo.nSys; i++) {
            const SysUploadTable *sysTable =
                (const SysUploadTable *)&uploadInfo.sysTables[i];
                
            if (*sysTable->enableState == SUBSYS_ENABLED ) {
                BIOMap *map = sysTable->bioMap[tid];
                if (map != NULL) {
                    int_T section;
                
                    intHdr[NSYS_IDX]++;

                    /* Add system index */
                    size = sizeof(int32_T);
                    overFlow = UploadBufAssignMem(
                        circBuf, size, &tmpHead, &bufMem);
                    if (overFlow) goto EXIT_POINT;
                    intHdr[NBYTES_IDX] += size;

                    CIRCBUF_COPY_DATA(bufMem, &i);

                    /* Add data values */
                    for (section=0; section<map->nSections; section++) {
                        BIOSection *sect = &map->sections[section];

                        overFlow = UploadBufAssignMem(
                            circBuf, sect->nBytes, &tmpHead, &bufMem);
                        if (overFlow) goto EXIT_POINT;
                        intHdr[NBYTES_IDX] += sect->nBytes;

                        CIRCBUF_COPY_DATA(bufMem, sect->start);
                    }
                }
            }
        }

        /* If no systems were active then, do nothing. */
        if (intHdr[NSYS_IDX] == 0) goto EXIT_POINT;

        /*
         * Go back and finish the header: [nBytes msgType nSys tid]
         */
        
        /* ...msgType */
        intHdr[MSG_TYPE_IDX] = UPMSG_DATA_PT;

        /* ...tid */
        intHdr[TID_IDX] = tid;
        CIRCBUF_COPY_DATA(msgStart, intHdr);

        /*
         * Time point successfully added to queue.
         */
        circBuf->head  = tmpHead;
        circBuf->empty = FALSE;

        if (preTrig) {
            trigInfo->preTrig.count++;
        }
    }

EXIT_POINT:
    if (!preTrig) {
        if (overFlow) {
            trigInfo->overFlow = TRUE;
            trigInfo->state    = TRIGGER_TERMINATING;
        }
#ifdef VXWORKS
        else if (trigInfo->state == TRIGGER_FIRED) {
            /* allow upload server to run - if data needs to be uploaded */
            semGive(uploadSem);
        }
#endif
    } 
} /* end UploadBufAddTimePoint */


/* Function ====================================================================
 * Send termination message for data data logging event on the first available
 * buffer.
 */
PRIVATE void UploadBufAddTerminator(void)
{
    int32_T     msg[2];
    int_T       tid;
    CircularBuf *circBuf     = NULL;
    const int_T NBYTES_IDX   = 0;
    const int_T MSG_TYPE_IDX = 1;

    /*
     * Find a buffer to use.
     */
    for (tid=0; tid<NUMST; tid++) {
        circBuf = &uploadInfo.circBufs[tid];
        if (circBuf->bufSize > 0) break;
    }

#if 0
#ifdef VERBOSE
    PRINT_M("\nSending data log termination flag on tid: %d\n", tid);
#endif
#endif

    msg[NBYTES_IDX] = (2*sizeof(int32_T));

    if (uploadInfo.trigInfo.holdOff == TRIGMODE_ONESHOT) {
        msg[MSG_TYPE_IDX] = UPMSG_TERMINATE_LOG_SESSION;
    } else {
        msg[MSG_TYPE_IDX] = UPMSG_TERMINATE_LOG_EVENT;
    }
    
    /* reset this buffer */
    circBuf->head = circBuf->buf;
    circBuf->tail = circBuf->buf;

    (void)memcpy(circBuf->head, msg, msg[NBYTES_IDX]);
    circBuf->head += msg[NBYTES_IDX];

    circBuf->empty   = FALSE;
    circBuf->newTail = NULL;
} /* end UploadBufAddTerminator */


/* Function ====================================================================
 * Called at the base rate, controls the state of data logging including:
 *   - monitoring the trigger signal for a trigger event
 *   - managing transition of most trigger states
 *      o a seperate function (UploadCheckEndTrigger manages the duration
 *        count and the transition from fired to terminating at the end
 *        of the data collection event).
 *
 * NOTE:
 *  o This function should be called after mdlOutputs for the base rate
 */
PUBLIC void UploadCheckTrigger(void)
{
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;

    if (trigInfo->state == TRIGGER_UNARMED) return;

    if (trigInfo->state == TRIGGER_HOLDING_OFF) {
        if (trigInfo->count++ == trigInfo->holdOff) {
            UploadArmTrigger();
        } else {
            return;
        }
    }

    /*
     * Transitions from the TRIGGER_ARMED_STATE to the TRIGGER_DELAYED
     * state or to the TRIGGER_FIRED_STATE are checked for and realized
     * within the task (tid) associated with the trigger signal.
     * See UploadBufAddTimePoint().
     */ 
    
    /*
     * Look for transitions from the TRIGGER_DELAYED state.  The TRIGGER_FIRED
     * state always follows the TRIGGER_DELAYED state.
     *
     * NOTE: the trigInfo count field is first used to count the trigger delay
     *       and then used to count the trigger duration
     */
    if (trigInfo->state == TRIGGER_DELAYED) {
        if (trigInfo->count++ >= trigInfo->delay) {
            trigInfo->count = trigInfo->preTrig.count; /* 0 unless pretrig */
            trigInfo->state = TRIGGER_FIRED;
            if (trigInfo->preTrig.duration > 0) {
                trigInfo->preTrig.checkUnderFlow = TRUE; /* xxx ? */
            }
#ifdef VERBOSE
            PRINT_M("\nTrigger fired!\n");
#endif
        }
    }
} /* end UploadCheckTrigger */


/* Function ====================================================================
 * Called at the base rate, controls the state of data logging wrt
 *  o incrementing the duration count
 *  o managing the transition to the trigger terminating state
 *
 * NOTES:
 *  o Call this function at the very end of a step.
 *  o Also see UploadCheckTrigger()
 */
PUBLIC void UploadCheckEndTrigger(void)
{
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;

    if (trigInfo->state == TRIGGER_UNARMED) return;

    /*
     * Increment duration count and terminate the data logging event if
     * the duration has been met.
     */
    if (trigInfo->state == TRIGGER_FIRED) {
        trigInfo->count++;
        if (trigInfo->count == trigInfo->duration) {
            trigInfo->state = TRIGGER_TERMINATING;
        }
    }

} /* end UploadCheckEndTrigger */


/* Function =======================================================================
 * Search through the upload buffers and fill out the internal copy of the
 * buffer list.  It contains a list of all buffer memory (1 entry per non-empty
 * tid buffer) that needs to be sent to the host.  Fill out the fields of the
 * specified ExtBufMemList (passed in by ext_svr) to provide public, read-only
 * access.
 */
PRIVATE void SetExtBufListFields(ExtBufMemList *extBufList)
{
    int_T       tid;
    BufMemList  *bufList  = &uploadInfo.bufMemList;

    bufList->nActiveBufs = 0;

    for (tid=0; tid<NUMST; tid++) {
        CircularBuf *circBuf = &uploadInfo.circBufs[tid];

        if (!circBuf->empty) {
            BufMem  *bufMem;
            char_T  *head   = circBuf->head;
            char_T  *tail   = circBuf->tail;
            int_T   size    = circBuf->bufSize;

            /* Validate that head/tail ptrs are within allocated range. */
            assert((head >= circBuf->buf) && (tail >= circBuf->buf));
            assert((head < circBuf->buf + circBuf->bufSize) &&
                   (tail < circBuf->buf + circBuf->bufSize));

            bufMem = &bufList->bufs[bufList->nActiveBufs];
            bufList->tids[bufList->nActiveBufs] = tid;
            assert(bufList->nActiveBufs < bufList->maxBufs);
            bufList->nActiveBufs++;

            bufMem->section1 = tail;
            circBuf->newTail = head;

            if (head > tail) {
                /* not wrapped - only one section required */
                bufMem->nBytes1  = head - tail;

                bufMem->nBytes2  = 0;
                bufMem->section2 = NULL;
            } else {
                /* wrapped - 2 sections required */
                bufMem->nBytes1 = circBuf->buf + size - tail;

                bufMem->nBytes2  = head - circBuf->buf;
                bufMem->section2 = circBuf->buf;
            }
        }
    }

    /*
     * Provide ext_svr with readonly access to the bufMemList.
     */
    extBufList->nActiveBufs = bufList->nActiveBufs;
    extBufList->bufs        = (const BufMem *)bufList->bufs;
    extBufList->tids        = (const int_T *)bufList->tids;
} /* end SetExtBufListFields */


/* Function =======================================================================
 * Set the internal copy of the buffer list to "empty" & fill out the fields of the
 * specified ExtBufMemList (passed in by ext_svr) to provide public, read only
 * access.
 */
PRIVATE void SetExtBufListFieldsForEmptyList(ExtBufMemList *extBufList)
{
    BufMemList *bufList  = &uploadInfo.bufMemList;

    bufList->nActiveBufs = 0;

    extBufList->nActiveBufs = bufList->nActiveBufs;
    extBufList->bufs        = (const BufMem *)NULL;
} /* end SetExtBufListFieldsForEmptyList */


/* Function ====================================================================
 * Called by ext_svr (background task), this function checks all buffers for
 * data and returns a list of buffer memory to be sent to the host.
 */
PUBLIC void UploadBufGetData(ExtBufMemList *extBufList)
{
    TriggerInfo *trigInfo = &uploadInfo.trigInfo;
 
    if ((trigInfo->state == TRIGGER_FIRED) ||
        (trigInfo->state == TRIGGER_TERMINATING)) {

        SetExtBufListFields(extBufList);

        /*
         * If all bufs are empty and we are terminating then we're now done!
         */
        if ((extBufList->nActiveBufs == 0) &&
            (trigInfo->state == TRIGGER_TERMINATING)) {
            
            trigInfo->state = TRIGGER_SENDING_TERM_MSG;

            /*
             * Send the final termination message.
             */
            UploadBufAddTerminator();

            /* Now we should find 1 buffer with data. */
            SetExtBufListFields(extBufList);
            assert(extBufList->nActiveBufs == 1);
        }
    } else {
        SetExtBufListFieldsForEmptyList(extBufList);
    }
    
} /* end UploadBufGetData */



/* [EOF] updown.c */
